
"""
              TuioDispatcher
                    |
                    |
            ---------------------
            |                   |
        TuioClient          TuioServer


"""


from typing import  Tuple


from pythonosc.udp_client import UDPClient
from pythonosc.osc_server import BlockingOSCUDPServer

from pythonosc.osc_message_builder import OscMessageBuilder
from pythonosc.osc_bundle_builder import OscBundleBuilder
from pythontuio.const import TUIO_BLOB, TUIO_CURSOR, TUIO_OBJECT
from pythontuio.dispatcher import TuioDispatcher




class TuioClient(TuioDispatcher, BlockingOSCUDPServer): # pylint: disable=too-many-ancestors
    """
    The TuioClient class is the central TUIO protocol decoder component.
    It provides a simple callback infrastructure using the TuioListener interface.
    In order to receive and decode TUIO messages an instance of TuioClient needs to be created.
    The TuioClient instance then generates TUIO events which are broadcasted to all
    registered classes that implement the TuioListener interface.
    """
    def __init__(self, server_address: Tuple[str, int]):
        TuioDispatcher.__init__(self)
        self._dispatcher = self
        BlockingOSCUDPServer.__init__(self,server_address, self)
        self.connected = False


    def server_bind(self):
        print(f"starting tuio-client at port {self.server_address[1]}")
        super().server_bind()
        self.handle_request()

    def server_close(self):
        print("stopping tuio-client")
        super().server_close()
    def start(self):
        """
        start serving for UDP OSC packages
        """
        self.serve_forever()

class TuioServer(TuioDispatcher, UDPClient):

    """
    Tuio client based on a basic osc udp client of the lib python-osc

    Notice that TuioSource is not implemented yet
    """

    def __init__(self, ip: str ="127.0.0.1" , port :int=3333):
        UDPClient.__init__(self,ip, port)
        TuioDispatcher.__init__(self)
        self._ip = ip
        self._port = port

        self.is_full_update : bool = False
        self._periodic_messages : bool = False
        self._intervall : int = 1000

    @staticmethod
    def _build_alive(address, profile_list):
        """
        builds a OSC which implements a alive message of TUIO
        returns the message
        """
        builder = OscMessageBuilder(address=address)
        builder.add_arg("alive")
        for profile in profile_list:
            builder.add_arg(profile.session_id) ## add id of cursors

        alive_msg = builder.build()
        return alive_msg
    def send_bundle(self):
        """Build :class:`OscMessage` from arguments and send to server

        Args:
            address: OSC address the message shall go to
            value: One or more arguments to be added to the message
        """

        bundle_builder = OscBundleBuilder(0)

        # build alive message
        if len(self.cursors) != 0:
            bundle_builder.add_content(TuioServer._build_alive(TUIO_CURSOR,self.cursors))
        if len(self.blobs) != 0:
            bundle_builder.add_content(TuioServer._build_alive(TUIO_BLOB,self.blobs))
        if len(self.objects) != 0:
            bundle_builder.add_content(TuioServer._build_alive(TUIO_OBJECT,self.objects))


        # set message of cursor
        for cursor in self.cursors:
            cursor_msg = cursor.get_message()
            bundle_builder.add_content(cursor_msg)

        # set message of blob
        for blob in self.blobs:
            blob_msg = blob.get_message()
            bundle_builder.add_content(blob_msg)

        # set message of object
        for o in self.objects:
            object_msg = o.get_message()
            bundle_builder.add_content(object_msg)

        # message fseq to end the bundle and send (optinal) frame id
        builder = OscMessageBuilder(address=TUIO_CURSOR)
        builder.add_arg("fseq")
        builder.add_arg(-1)
        fseq = builder.build()
        bundle_builder.add_content(fseq)

        # build bundle and send
        bundle = bundle_builder.build()
        self.send(bundle)

    def disable_periodic_messages(self, ):
        """
        Not implemented
        """
        self._periodic_messages = False
        raise Exception("Not implemented")

    def enable_periodic_messages(self, intervall:int):
        """
        Not implemented
        """
        self._periodic_messages = True
        self._intervall = intervall
        raise Exception("Not implemented")

    def set_source_name(self, _name : str, ip :str=None):
        """
        Not implemented
        """
        if ip is not None:
            ip = self._ip

        raise Exception("Not implemented")




"""
this python file is written orientated by the TUIO spezification
https://www.tuio.org/?specification
It supports only 2D Object|Blob|Cursor

            Profile
                |
    ---------------------
    |           |       |
  Object     Cursor    Blob


"""

from pythonosc.osc_message_builder import OscMessageBuilder
from pythonosc.osc_message import OscMessage

from pythontuio.const import TUIO_BLOB, TUIO_CURSOR, TUIO_OBJECT

class Profile:
    """
    custom class of all subjects passing the TUIO connection.
    See more at https://www.tuio.org/?specification

    """

    def __init__(self, session_id):
        self.session_id = session_id

class Object(Profile):
    """
    TUIO Object 2D Interactive Surface
    """
    def __init__(self, session_id):
        super(Object, self).__init__(session_id)
        self.class_id               = -1            # i
        self.position               = (0, 0)   # x,y
        self.angle                  = 0             # a
        self.velocity               = (0, 0)   # X,Y
        self.velocity_rotation      = 0             # A
        self.motion_acceleration    = 0             # m
        self.rotation_acceleration  = 0             # r

    def get_message(self) -> OscMessage:
        """
        returns the OSC message of the Object with the TUIO spezification
        """
        x, y = self.position
        X, Y = self.velocity
        builder = OscMessageBuilder(address=TUIO_OBJECT)
        for val in [
                "set",
                int(self.session_id),
                int(self.class_id),
                float(x),
                float(y),
                float(self.angle),
                float(X),
                float(Y),
                float(self.velocity_rotation),
                float(self.motion_acceleration),
                float(self.rotation_acceleration)
        ]:
            builder.add_arg(val)
        return builder.build()

class Cursor(Profile):
    """
    TUIO Cursor 2D Interactive Surface
    """
    def __init__(self, session_id):
        super(Cursor, self).__init__(session_id)
        self.position               = (0, 0)   # x,y
        self.velocity               = (0, 0)   # X,Y
        self.motion_acceleration    = 0             # m

    def get_message(self)-> OscMessage:
        """
        returns the OSC message of the Cursor with the TUIO spezification
        """
        x, y = self.position
        X, Y = self.velocity
        builder = OscMessageBuilder(address=TUIO_CURSOR)
        for val in [
                "set",
                self.session_id,
                float(x),
                float(y),
                float(X),
                float(Y),
                float(self.motion_acceleration)
        ]:
            builder.add_arg(val)
        return builder.build()


class Blob(Profile):
     # pylint: disable=too-many-instance-attributes
    """
    TUIO Blob 2D Interactive Surface
    """
    def __init__(self, session_id):
        super(Blob, self).__init__(session_id)
        self.position               = (0, 0)        # x,y
        self.angle                  =  5            # a
        self.dimension              = (.1, .1)      # w, h
        self.area                   = 0.1           # f
        self.velocity               = (0.1, 0.1)    # X,Y
        self.velocity_rotation      = 0.1           # A
        self.motion_acceleration    = 0.1           # m
        self.rotation_acceleration  = 0.1           # r

    def get_message(self)-> OscMessage:
        """
        returns the OSC message of the Blob with the TUIO spezification
        """
        x, y = self.position
        X, Y = self.velocity
        w, h = self.dimension
        builder = OscMessageBuilder(address=TUIO_BLOB)
        for val in [
                "set",
                self.session_id,
                float(x),
                float(y),
                float(self.angle),
                float(w),
                float(h),
                float(self.area),
                float(X),
                float(Y),
                float(self.velocity_rotation),
                float(self.motion_acceleration),
                float(self.rotation_acceleration)
        ]:
            builder.add_arg(val)
        return builder.build()



"""
init file of the pip3 module pythontuio
"""
from pythontuio.tuio_profiles import Cursor
from pythontuio.tuio_profiles import Blob
from pythontuio.tuio_profiles import Object

from pythontuio.tuio import TuioServer
from pythontuio.tuio import TuioClient
from pythontuio.dispatcher import TuioListener




"""
File for constant TUIO values
"""
TUIO_CURSOR = "/tuio/2Dcur"
TUIO_OBJECT = "/tuio/2Dobj"
TUIO_BLOB   = "/tuio/2Dblb"

TUIO_ALIVE = "alive"
TUIO_SET =  "set"
TUIO_END = "fseq"
TUIO_SOURCE = "source"



"""
classes to handle incoming osc messages
"""
from abc import ABC # abstract base class of python

from pythonosc.dispatcher import Dispatcher

from pythontuio.tuio_profiles import Cursor, Blob, Object
from pythontuio.tuio_profiles import TUIO_BLOB, TUIO_CURSOR, TUIO_OBJECT

from pythontuio.const import TUIO_END,TUIO_ALIVE,TUIO_SET, TUIO_SOURCE



# pylint: disable=unnecessary-pass
class TuioListener(ABC):
    """
    Abstract TuioListener to define callbacks für the diffrent tuio events
    """
    def add_tuio_object(self, obj):
        """Abstract function to add a behavior for tuio add object event"""
        pass
    def update_tuio_object(self, obj):
        """Abstract function to add a behavior for tuio update object event"""
        pass
    def remove_tuio_object(self, obj):
        """Abstract function to add a behavior for tuio remove object event"""
        pass

    def add_tuio_cursor(self, cur):
        """Abstract function to add a behavior for tuio add cursor event"""
        pass
    def update_tuio_cursor(self, cur):
        """Abstract function to add a behavior for tuio update cursor event"""
        pass
    def remove_tuio_cursor(self, cur):
        """Abstract function to add a behavior for tuio remove cursor event"""
        pass

    def add_tuio_blob(self, blob):
        """Abstract function to add a behavior for tuio add blob event"""
        pass
    def update_tuio_blob(self, blob):
        """Abstract function to add a behavior for tuio update blob event"""
        pass
    def remove_tuio_blob(self, blob):
        """Abstract function to add a behavior for tuio remove blob event"""
        pass
    def refresh(self, time):
        """Abstract This callback method is invoked by the TuioClient
        to mark the end of a received TUIO message bundle."""
        pass
# pylint: enable=unnecessary-pass


class TuioDispatcher(Dispatcher):
    """
    class to hold Eventlistener and the TuioCursors, TuioBlobs, and TuioObjects
    """
    def __init__(self):
        super(TuioDispatcher, self).__init__()
        self.cursors : list = []
        self.objects : list = []
        self.blobs   : list = []
        self._listener : list = []
        self.map(f"{TUIO_CURSOR}*", self._cursor_handler)
        self.map(f"{TUIO_OBJECT}*", self._object_handler)
        self.map(f"{TUIO_BLOB}*", self._blob_handler)
        self.set_default_handler(self._default_handler)

    def _cursor_handler(self, address, *args):
        """
        callback to convert OSC message into TUIO Cursor
        """
        if len(args) == 0 :
            raise Exception("TUIO message is Broken. No TUIO type specified")
        ttype = args[0]
        args = args[1:]
        print(f"{address}:{ttype} {args}")
        if ttype == TUIO_ALIVE :
            cursors = self.cursors.copy()
            self.cursors = _sort_matchs(cursors,args,Cursor)

        elif ttype == TUIO_SET:
            for cursor in self.cursors:
                if cursor.session_id != args[0]:
                    continue
                cursor.position = (args[1], args[2])
                cursor.velocity = (args[3], args[4])
                cursor.motion_acceleration = args[5]
                # call add cursor event
                for listener in self._listener:
                    listener.add_tuio_cursor(cursor)


        elif ttype == TUIO_END:
            return # nothing to happend here
        elif ttype == TUIO_SOURCE:
            print(f"Message by {args} reveiced")
        else:
            raise Exception("Broken TUIO Package")


    def _object_handler(self, address, *args):
        """
        callback to convert OSC message into TUIO Object
        """
        if len(args) == 0 :
            raise Exception("TUIO message is Broken. No TUIO type specified")
        ttype = args[0]
        args = args[1:]
        print(f"{address}:{ttype} {args}")
        print("ttype !!!!!!!!!!!!!!!!!!!!!{ttype}")
        if ttype == TUIO_ALIVE :
            objects = self.objects.copy()
            self.cursors = _sort_matchs(objects,args,Object)

        elif ttype == TUIO_SET:
            for obj in self.objects:
                if obj.session_id != args[0]:
                    continue
                obj.class_id               = args[1]                # i
                obj.position               = (args[2], args[3])     # x,y
                obj.angle                  = args[4]                # a
                obj.velocity               = (args[5], args[6])     # X,Y
                obj.velocity_rotation      = args[7]                # A
                obj.motion_acceleration    = args[8]                # m
                obj.rotation_acceleration  = args[9]                # r

                 # call add object event
                map(lambda x, obj=obj: x.add_tuio_object(obj),self._listener)
        elif ttype == TUIO_END:
            return # nothing to happend here
        elif ttype == TUIO_SOURCE:
            print(f"Message by {args} reveiced")
        else:
            raise Exception("Broken TUIO Package")

    def _blob_handler(self, address, *args):
        """
        callback to convert OSC message into TUIO Blob
         """
        if len(args) == 0 :
            raise Exception("TUIO message is Broken. No TUIO type specified")
        ttype = args[0]
        args = args[1:]
        print(f"{address}:{ttype} {args}")
        if ttype == TUIO_ALIVE :
            blobs = self.blobs.copy()
            self.blobs = _sort_matchs(blobs,args,Blob)

        elif ttype == TUIO_SET:
            for blob in self.blobs:
                if blob.session_id != args[0]:
                    continue
                blob.position               = (args[1], args[2])     # x,y
                blob.angle                  = args[3]                # a
                blob.dimension              = (args[4], args[5])     # w, h
                blob.area                   = args[6]                # f
                blob.velocity               = (args[7], args[8])     # X,Y
                blob.velocity_rotation      = args[9]                # A
                blob.motion_acceleration    = args[10]               # m
                blob.rotation_acceleration  = args[11]               # r
                # call add blob event
                map(lambda x, blob=blob: x.add_tuio_blob(blob),self._listener)

        elif ttype == TUIO_END:
            return # nothing to happend here
        elif ttype == TUIO_SOURCE:
            print(f"Message by {args} reveiced")
        else:
            raise Exception("Broken TUIO Package")

    def add_listener(self, listener :TuioListener):
        """
        Adds the provided TuioListener to the list of registered TUIO event listeners
        """
        self._listener.append(listener)

    def remove_listener(self, listener :TuioListener):
        """
        Removes the provided TuioListener from the list of registered TUIO event listeners
        """
        self._listener.remove(listener)

    def remove_all_listeners(self):
        """
        Removes all provided TuioListeners from the list of registered TUIO event listeners
        """
        self._listener.clear()
def _sort_matchs(profiles, session_ids, Profile_type):
    new_profiles = []
    for session_id in session_ids:
        profil_match = None
        # search for cursor
        for profile in profiles:
            if profile.session_id == session_id:
                profil_match = profile
        # if match found copy it
        if profil_match is not None:
            new_profiles.append(profil_match)
        # else add new one
        else :
            new_profiles.append(Profile_type(session_id))
    return new_profiles
